name = 'igittigitt'
title = 'A spec-compliant gitignore parser for Python'
version = 'v1.0.2'
url = 'https://github.com/bitranox/igittigitt'
author = 'Robert Nowotny'
author_email = 'bitranox@gmail.com'
shell_command = 'igittigitt'


def print_info() -> None:
    print("""\

Info for igittigitt:

    A spec-compliant gitignore parser for Python

    Version : v1.0.2
    Url     : https://github.com/bitranox/igittigitt
    Author  : Robert Nowotny
    Email   : bitranox@gmail.com""")
